/*
 * File : frame.h
 *
 * Description : This file contains constant definitions and
 *	the frame class declaration.
 */

#ifndef _frame_h_
#define _frame_h_

#include "types.h"
#include "list.h"

/*
 * Some flags, and masks.
 */


unsigned const char FLAG = 0x7e; 		 

const char	MASK		= 0x01;
const char 	SUP_MASK	= 0x0f;
const char 	UNUM_MASK	= 0xef;

const char	PF_MASK = 0x10; 
const char 	NS_MASK = 0x0e;
const char 	NR_MASK = 0xe0;


/*
 * Literal values for frames
 */

const unsigned char IFRAME 	= 0x00;
const unsigned char RR 		= 0x01;
const unsigned char RNR 	= 0x05;
const unsigned char REJ 	= 0x09;
const unsigned char SABM 	= 0x2f;
const unsigned char DISC 	= 0x43;
const unsigned char UA 		= 0x63;
const unsigned char DM		= 0x0f;
const unsigned char FRMR 	= 0x87;

/*
 * Some error conditions
 */

// an undefined frame
const unsigned char UNDEFINED		= 0xaa; 
// an invalid frame
const unsigned char INVALID	    	= 0xbb; 
// Info in I-frame exceeds max.
const unsigned char IfrLengthExcess = 0xcc; 
// Frame has invalid length.
const unsigned char IncorrectLength = 0xdd; 

/*
 * FCS Defines
 */

const unsigned short FCS_INIT = 0xffff;
const unsigned short FCS_RECV = 0xf0b8;

class dataPacket;
class frame;
class dllayer;

/*
 * Define frameMember and frameList - since
 * frame objects will definitely be in a list
 */
DefineListMember(frame);
DefineList(frame);


/*
 * Class : frame
 *
 * Description : The frame class encapsulates all the 
 * functionality needed to form data-link frames for 
 * transmission, and parse incoming data-link frames to 
 * determine the frame parameters like n(r), n(s), p/f 
 * fields, and whether its a command or response, and of 
 * course the frame type. The class also has the ability to
 * calculate frame check sequences (FCS) for frames to
 * be transmitted, and verify FCS for received frames.
 * Methods are provided to access frame parameters, and also 
 * to form frames for transmission.
 *
 * Attributes :
 *	dataPacket *_data : Pointer to the data-packet which holds
 *		data.
 *	unsigned char _cntl : The control byte for this frame.
 *	unsigned char _ft: The frame type.
 *	unsigned int _pf : The P/F bit
 *	unsigned int _ns : The N(s) value
 *	unsigned int _nr : The N(r) value
 *	Boolean _cmd : True means this frame is a COMMAND frame. 
 *	Boolean _formed : True means this frame has already been
 *		formed (i.e. flags, address, control bytes etc. are 
 *		in place.)
 */

class frame : public frameMember {

 public:
	frame();
	frame(dataPacket *aData, Boolean xmit);
	~frame();

	void SetParams(unsigned char ft, int pf, int nr = 0, int ns = 0) 
		{ _pf = pf; _ns = ns; _nr = nr; _ft = ft; }

	void FormFrame(Boolean isCmd);

	dataPacket *	GetDataPacket() { return _data; }
	unsigned char 	GetFrameType() { return _ft; }

	Boolean IsCmd() { return _cmd; }
	unsigned int GetNr() { return _nr; }
	unsigned int GetNs() { return _ns; }
	unsigned int GetPf() { return _pf; }
	unsigned char GetCntl() { return _cntl; }
	void SetNs(int ns) { _ns = ns; }
	void SetNr(int nr) { _nr = nr; }
	void SetPF(int pf) { _pf = pf; }

	void 	PrintParams();

 protected:

	void 	ExtractParams();
	void 	AppendFCS();
	Boolean CheckValid();
	unsigned short CalculateCksum(unsigned short crc,
			const unsigned char *s, int n);

	dataPacket *_data;
	unsigned char _cntl;
	unsigned char _ft;
	unsigned int _pf : 1;
	unsigned int _ns : 3;
	unsigned int _nr : 3;
	Boolean _cmd;
	Boolean _formed;
};

#endif
