/*
 * File : nwlayer.C
 *
 * Description: This file contains the method definitions for the
 *	nwlayer class.
 */

#include <iostream.h>
#include <stdio.h>
#include <unistd.h>
#include <osfcn.h>
#include <string.h>
#include "scheduler.h"
#include "timerHandler.h"
#include "trace.h"
#include "dataPacket.h"
#include "nwlayer.h"
#include "phEvents.h"
#include "dlEvents.h"
#include "errorCodes.h"
#include "frame.h"
#include "dlCmdResTimer.h"

DEFINE_TRACE_MODULE(nwlayerModule,"nwlayerModule");

/*
 * System parameters - these can be changed by the user.
 */

int T1 = 10;
int N2 = 10;
unsigned char cmdAddress;
unsigned char resAddress;

unsigned const char errorAddress = 0xff;

unsigned const char ERROR_FLAG = 0xee;

/*
 * Receiver busy flag. The data-link looks at
 * this flag to simulate a receiver busy condition.
 * This can be set from the UI.
 */

Boolean receiverBusy = FALSE;

/*
 * Maximum sizes for packets - this can also
 * be set from the UI>
 */

int MaxIframeLength = 200;
int MaxPacketSize = 210;

/*
 * Function : nwlayer(int aFd)
 *
 * Description: This is the constructor. It calls the base class
 * constructors, and invokes PrintCommands() to print out
 * the top-level menu.
 */

nwlayer::nwlayer(int aFd) : inputHandler (aFd),
_acceptConnReq(TRUE)
{
	PrintCommands();
}

/*
 * Function : ~nwlayer()
 *
 * Description : This is the destructor. Does nothing.
 */

nwlayer::~nwlayer()
{
}

/*
 * Function : EventCallback()
 *
 * Description : This function is redefined to handle input from
 * STDIN. A line of user input is read, and the function Parse()
 * is called.
 */

void 
nwlayer::EventCallback()
{
	char buf[MAXLINE];

	TRACE(nwlayerModule,1,("\nEntering nwlayer::EventCallback\n"));


	int bytes;
	if ( (bytes = read(_fd,buf,MAXLINE))  == -1 ) {
		perror("Error in read");
		return;
	}
	buf[bytes] = '\0';
	TRACE(nwlayerModule,1,("\nRead : %s\n",buf));
	Parse(buf);
	TRACE(nwlayerModule,1,("Exiting nwlayer::EventCallback\n"));
}

/*
 * Function : Parse(char *buf)
 *
 * Description : This function figures out the user-input for
 * the top-level menu and invokes the appropriate function.
 */

void
nwlayer::Parse(char *buf)
{
	int cmd = atoi(buf);

	TRACE(nwlayerModule,1,("\nCommand : %d\n",cmd));

	switch (cmd ) {
		
		case 1:
			Trace(TRUE);			
			break;
		case 2:
			Trace(FALSE);
			break;
		case 3:
			PhysicalLayerCommands();
			break;
		case 4:
			DataLinkLayerCommands();
			break;
		case 5:
			NetworkLayerCommands();
			break;
		case 6:
			SetSystemParams();
			break;
		case 7:
			exit(0);
			break;
		default:
			cout << "Unknown command\n" << flush;
	}

	PrintCommands();
}


/*
 * Function : PrintCommands()
 *
 * Description : This function prints out the top-level 
 * menu.
 */

void
nwlayer::PrintCommands()
{
	cout << "\nThe following commands are available\n";
	cout << "Use the integer in brackets to invoke the given command\n";

	cout << "\n";
	cout << "Set TRACE               (1)\n";
	cout << "Unset TRACE             (2)\n";
	cout << "Physical Layer commands (3)\n";
	cout << "Data link layer commands(4)\n";
	cout << "Network layer commands  (5)\n";
	cout << "Set system parameters   (6)\n";
	cout << "Exit                    (7)\n";
	cout << "Please type in (integer) command : " << flush;
}

/*
 * Function : Trace(Boolean set)
 *
 * Description : This function provides the means to Set/Reset
 * the tracing on any given module.
 */

void
nwlayer::Trace(Boolean set)
{
	int module, level;

	cout << "\n";
	traceManager::PrintModules();
	cout << "Type Module :";
	cin >> module;

	trace *info = traceManager::GetModule(module);
	if ( info == 0 )
		return;

	if ( set == TRUE ) {
		cout << "Type Level :";
		cin >> level;
		info->SetTrace(level);
	}
	else
		info->ResetTrace();
}

/*
 * Function : PhysicalLayerCommands()
 *
 * Description : This function provides a menu-driven interface
 * to interact with the physical layer directly. 
 */

void
nwlayer::PhysicalLayerCommands()
{
	int cmd;
	unsigned char msg[100];

	cout << "\n              Activate Physical Layer      (1)";
	cout << "\n              Deactivate Physical Layer    (2)";
	cout << "\n              Send Data                    (3)";
	cout << "\n              Block transmission           (4)";
	cout << "\n              Block reception              (5)";
	cout << "\n              Unblock transmission         (6)";
	cout << "\n              Unblock reception            (7)";
	cout << "\n              \nType in command :" << flush;
	cin >> cmd;

	layer *ph = _nMinus1->GetNMinus1Layer();
	int errorCode = codeSuccessful;
	switch (cmd ) {
		case 1:
			ph->Command(PH_ActivateReq,errorCode);
			break;
		case 2:
			ph->Command(PH_DeactivateReq,errorCode);
			break;
		case 3:
			cout << "\n       Type in message: " << flush;
			cin >> msg;
			dataPacket *pkt = new dataPacket(msg,strlen((char *)msg));
			pkt->Register();
			pkt->Prepend((unsigned char *)"Header ",7);
			pkt->Append((unsigned char *)" Trailer",8);
			ph->Command(PH_DataReq,errorCode,pkt);
			pkt->Deregister();
			break;
		case 4:
			ph->BlockXmit();
			break;
		case 5:
			ph->BlockRecv();
			break;
		case 6:
			ph->UnblockXmit();
			break;
		case 7:
			ph->UnblockRecv();
			break;
		default:
			cout << "Unknown command\n";
	}
	if ( errorCode != codeSuccessful )
		cout << "Error return : " << errorCode << "\n";
}

/*
 * Function : DataLinkLayerCommands()
 *
 * Description : This function provides a menu-driven interface
 * to interact with the data-link layer. Valid and invalid frames
 * can also be sent to the remote DTE while bypassing the local
 * DTE entirely.
 */

void
nwlayer::DataLinkLayerCommands()
{
	dataPacket *data;
	frame *fr;
	int errorCode = codeSuccessful; 
	layer *ph = _nMinus1->GetNMinus1Layer();

	cout << "  Send valid frame - bypass data link (0)\n";
	cout << "  Send invalid frame-bypass data link (1)\n"; 
	cout << "  Data Link Connect Req               (2)\n"; 
	cout << "  Data Link Connect Res               (3)\n"; 
	cout << "  Data Link Data Req                  (4)\n";
	cout << "  Data Link Reset   Req               (5)\n"; 
	cout << "  Data Link Reset   Res               (6)\n"; 
	cout << "  Data Link Disconnect Req            (7)\n"; 
	cout << "  Block transmission                  (8)\n";
	cout << "  Block reception                     (9)\n";
	cout << "  Unblock transmission                (10)\n";
	cout << "  Unblock reception                   (11)\n";
	cout << "  Set Data Link to busy               (12)\n";
	cout << "  Set Data Link to ready              (13)\n";


	cout << "  Type in command : " << flush;

	int cmd;
	cin >> cmd;

	switch ( cmd ) 
		{
	case 0:
		data = GetInfoField();
		fr = new frame(data,TRUE);
		SetFrameType(fr);
		fr->PrintParams();
		ph->Command(PH_DataReq,errorCode,fr->GetDataPacket());
		if ( errorCode != codeSuccessful )
			cout << "Error return " << errorCode << "\n";
		delete fr;
		break;
	case 1:
		data = GetInfoField();
		fr = new frame(data,TRUE);
		SetFrameType(fr);

		cout << "        Mess up flags                        (1)\n";
		cout << "        Mess up FCS                          (2)\n";
		cout << "        Change address encoding              (3)\n";
		cout << "        Have less than 32 bits between flags (4)\n";
		cout << "        Toggle command / response frmames    (5)\n";
		cout << "\n        Select type of error to be introduced: ";
		cin >> cmd;
		switch (cmd ) {
			case 1:
				MessUpFlags(fr);
				break;
			case 2:
				MessUpFCS(fr);
				break;
			case 3:
				ChangeAddress(fr);
				break;
			case 4:
				LessThan32Bits(fr);
				break;
			case 5:
				fr->FormFrame( fr->IsCmd() == TRUE ? FALSE : TRUE );
				break;
			default:
				cout << "Unknown command\n";
				break;
		}

		fr->PrintParams();
		ph->Command(PH_DataReq,errorCode,fr->GetDataPacket());
		if ( errorCode != codeSuccessful )
			cout << "Error return " << errorCode << "\n";
		delete fr;
		break;
	case 2:
		if (_nMinus1)
			_nMinus1->Command(DL_ConnectReq,errorCode);
		break;
	case 3:
		if (_nMinus1)
			_nMinus1->Command(DL_ConnectRes,errorCode);
		break;
	case 4:
		int num = 0;
		cout << "Number of requests: " << flush;
		cin >> num;
		
		unsigned char msg[100];
		cout << "\n       Type in message: " << flush;
		cin >> msg;
		if (_nMinus1)
			while ( num-- ) {
				dataPacket *pkt = new dataPacket(msg,strlen((char *)msg));
				pkt->Register();
				_nMinus1->Command(DL_DataReq,errorCode,pkt);
				pkt->Deregister();
			}
		break;
	case 5:
		if (_nMinus1)
			_nMinus1->Command(DL_ResetReq,errorCode);
		break;
	case 6:
		if (_nMinus1)
			_nMinus1->Command(DL_ResetRes,errorCode);
		break;
	case 7:
		if (_nMinus1)
			_nMinus1->Command(DL_DisconnectReq,errorCode);
		break;
	case 8:
		if (_nMinus1 )
			_nMinus1->BlockXmit();
		break;
	case 9:
		if (_nMinus1 )
			_nMinus1->BlockRecv();
		break;
	case 10:
		if (_nMinus1 )
			_nMinus1->UnblockXmit();
		break;
	case 11:
		if (_nMinus1 )
			_nMinus1->UnblockRecv();
		break;
	case 12:
			receiverBusy = TRUE;
			break;
	case 13:
			receiverBusy = FALSE;
			break;
	default:
		cout << "      Unknown command\n" << flush;
		break;
		}
}

/*
 * Function : GetPfBit()
 *
 * Description: This function queries the user for a
 * value of the P/F bit and returns the value.
 */

int
nwlayer::GetPfBit()
{
	int pf;
	cout << "                  Type in P/F bit :";
	cin >> pf;
	return pf;
}

/*
 * Function : IsCmdRes()
 *
 * Description: This function queries the user for a
 * value the type of frame (Command or Response) and
 * returns the value.
 */

Boolean 
nwlayer::IsCmdRes()
{
	int cmdRes;
	cout << "                  Command (1) or Response (0) :";
	cin >> cmdRes;
	return ((Boolean) cmdRes);
}

/*
 * Function : GetNr()
 *
 * Description: This function queries the user for a
 * value of the N(r) field and returns the value.
 */

int		
nwlayer::GetNr()
{
	int nr;
	cout << "                  Type in NR value :";
	cin >> nr;
	return nr;
}

/*
 * Function : GetNs()
 *
 * Description: This function queries the user for a
 * value of the N(s) field and returns the value.
 */

int		
nwlayer::GetNs()
{
	int ns;
	cout << "                  Type in NS value :";
	cin >> ns;
	return ns;
}

/*
 * Function : Command(int cmd, int& errorCode)
 *
 * Description: This function implements the data-link
 * primitives DL_ConnectInd, DL_ConnectCnf, DL_ResetCnf
 * The flag _acceptConnReq is examined before accepting
 * or rejecting a connection request.
 */

void 
nwlayer::Command(int cmd, int& errorCode)
{
	scheduler *sch = scheduler::GetScheduler();
	int res;

	if ( _blockRecv == TRUE )
		return;

	switch(cmd) 
		{
	case DL_ConnectInd:
		(_acceptConnReq == TRUE ) 
			? (res = DL_ConnectRes) : (res = DL_DisconnectReq); 

		// Start a timer - it will respond to the dllayer
		// when it expires.
		sch->AddEventHandler(new dlCmdResTimer(res,_nMinus1));

		TRACE(nwlayerModule,2,("Received Connect Request\n"));
		errorCode = codeSuccessful;
		break;
	case DL_ConnectCnf:
		TRACE(nwlayerModule,2,("Connection Established\n"));
		errorCode = codeSuccessful;
		break;
	case DL_ResetCnf:
		TRACE(nwlayerModule,2,("Connection Reset\n"));
		errorCode = codeSuccessful;
		break;
	default:
		TRACE(nwlayerModule,2,("Unknown command\n"));
		errorCode = codeGeneralFailure;
		break;
		}
}

void 
nwlayer::Command(int cmd, int &errorCode, dataPacket *data) 
{
}

/*
 * Function : Command(int cmd, int &errorCode, int reason)
 * 
 * Description : This function implements the DL_ResetInd,
 * DL_DisconnectInd primitives. The flag _acceptConnReq is
 * examined before a Reset request is accepted or rejected.
 */


void 
nwlayer::Command(int cmd, int &errorCode, int reason) 
{
	if ( _blockRecv == TRUE )
		return;

	int res;
	scheduler *sch = scheduler::GetScheduler();

	switch(cmd)
		{
	case DL_ResetInd:
		TRACE(nwlayerModule,2,("Reset Indication : %d\n",reason));
		if ( reason == internalError)
			return; // no need for a reset response

		(_acceptConnReq == TRUE ) 
			? (res = DL_ResetRes) : (res = DL_DisconnectReq); 
		sch->AddEventHandler(new dlCmdResTimer(res,_nMinus1));
		TRACE(nwlayerModule,2,("Received Reset Request\n"));
		errorCode = codeSuccessful;
		break;
	case DL_DisconnectInd:
		TRACE(nwlayerModule,2,("Disconnected, reason : %d\n",reason));
		errorCode = codeSuccessful;
		break;
	default:
		TRACE(nwlayerModule,2,("Unknown command\n"));
		errorCode = codeGeneralFailure;
		break;
		}
}


/*
 * Function : NetworkLayerCommands()
 *
 * Description : This function provides a menu-driven interface
 * to the network layer.
 */

void
nwlayer::NetworkLayerCommands()
{
	int cmd;

	cout << "\n              Block reception              (1)";
	cout << "\n              Unblock reception            (2)";
	cout << "\n              NW conn accept toggle        (3)";
	cout << "\n              \nType in command :" << flush;

	cin >> cmd;

	switch (cmd ) {
		case 1:
			BlockRecv();
			break;
		case 2:
			UnblockRecv();
			break;
		case 3:
			if ( _acceptConnReq == TRUE )
				_acceptConnReq = FALSE;
			else
				_acceptConnReq = TRUE;
			break;
		default:
			cout << "Unknown command\n";
			break;
	}
}

/*
 * Function : SetSystemParams()
 *
 * Description : This function provides a menu-driven interface
 * to set systems parameters.
 */

void
nwlayer::SetSystemParams()
{
	cout << "\n         Timer T1                (1)";
	cout << "\n         Retry count N2          (2)";
	cout << "\n         Maximum I-frame length  (3)";
	cout << "\n         Maximum data packet size(4)";
	cout << "\n         Please select : ";
	int attr;
	cin >> attr;
	cout << "\n         Value : ";

	switch (attr) 
		{
	case 1:
		cin >> T1;
		break;
	case 2:
		cin >> N2;
		break;
	case 3:
		cin >> MaxIframeLength;
		break;
	case 4:
		cin >> MaxPacketSize;
		break;
	default:
		cout << "Unknown parameter\n";
		break;
	}
}

/*
 * Function : SetFrameType(frame *fr)
 *
 * Description : This function takes a pointer to a frame,
 * and sets to a user specified frame type, with the
 * the appropriate values for n(r),n(s),p/f etc.
 */

void
nwlayer::SetFrameType(frame *fr)
{
			
	cout << "                  IFRAME ( 0  )\n";
	cout << "                  RR     ( 1  )\n";
	cout << "                  RNR    ( 5  )\n";
	cout << "                  REJ    ( 9  )\n";
	cout << "                  DM     ( 15 )\n";
	cout << "                  SABM   ( 47 )\n";
	cout << "                  DISC   ( 67 )\n";
	cout << "                  UA     ( 99 )\n";
	cout << "                  FRMR   ( 135)\n";
	cout << "    Unknown control field( 13 )\n";
	cout << "                  Select frame type : " << flush;

	unsigned int frameType;
	cin >> frameType;


	switch ( frameType ) {
	case IFRAME:
		fr->SetParams(IFRAME,GetPfBit(),GetNr(),GetNs());
		fr->FormFrame(TRUE);
		break;
	case RR:
		fr->SetParams(RR,GetPfBit(),GetNr());
		fr->FormFrame(IsCmdRes());
		break;
	case RNR:
		fr->SetParams(RNR,GetPfBit(),GetNr());
		fr->FormFrame(IsCmdRes());
		break;
	case REJ:
		fr->SetParams(REJ,GetPfBit(),GetNr());
		fr->FormFrame(IsCmdRes());
		break;
	case SABM:
		fr->SetParams(SABM,GetPfBit());
		fr->FormFrame(TRUE);
		break;
	case DISC:
		fr->SetParams(DISC,GetPfBit());
		fr->FormFrame(TRUE);
		break;
	case UA:
		fr->SetParams(UA,GetPfBit());
		fr->FormFrame(FALSE);
		break;
	case DM:
		fr->SetParams(DM,GetPfBit());
		fr->FormFrame(FALSE);
		break;
	case FRMR:
		fr->SetParams(FRMR,GetPfBit());
		fr->FormFrame(FALSE);
		break;
	case 0x0d:
		fr->SetParams(0x0d,GetPfBit());
		fr->FormFrame(IsCmdRes());
		break;
	default:
		cout << "                 Unknown frame type :" << frameType;
		cout << flush;
		return;
	}
}

/*
 * Function : GetInfoField()
 *
 * Description: This function queries the user for any
 * information field that needs to sent in an Iframe.
 * Returns a dataPacket containing the info.
 */

dataPacket *
nwlayer::GetInfoField()
{
	cout << "\nDo you want any information field (y/n) : ";
	char yesNo;
	dataPacket *data;
	cin >> yesNo;
	if ( yesNo == 'y') {
		unsigned char msg[50];
		cout << "Type in information (no spaces): ";
		cin >>  msg;
		data = new dataPacket(msg,strlen((char *)msg));	
	}
	else
		data = new dataPacket(); // create an empty pkt.
	return data;
}

/*
 * Function : MessUpFlags(frame *fr)
 *
 * Description : This function takes a pointer to a frame
 * and messes up the flags in the frame.
 */

void
nwlayer::MessUpFlags(frame *fr)
{
	cout << "Appending non-flag byte to frame\n";
	dataPacket *data = fr->GetDataPacket();
	data->Append((unsigned char *)&ERROR_FLAG,1); 
}

/*
 * Function : MessUpFCS(frame *fr)
 *
 * Description : This function takes a pointer to a frame
 * and messes up the FCS in the frame.
 */

void
nwlayer::MessUpFCS(frame *fr)
{
	cout << "Adding an extra byte..should mess it up!\n";
	dataPacket *data = fr->GetDataPacket();
	data->PostStrip(1); // remove flag
	data->Append((unsigned char *)"X",1);
	data->Append((unsigned char *)&FLAG,1); // Put Flag back..
}

/*
 * Function : LessThan32Bits(frame *fr)
 *
 * Description : This function takes a pointer to a frame
 * and ensures that the frame has less than 32 bits between
 * flags.
 */

void
nwlayer::LessThan32Bits(frame *fr)
{
	cout << "Making frame less than 32 bits\n";
	dataPacket *data = fr->GetDataPacket();
	data->PreStrip(2);
	data->Prepend((unsigned char *)&FLAG,1); // put flag back
}

/*
 * Function : ChangeAddress(frame *fr)
 *
 * Description : This function takes a pointer to a frame
 * and changes the address field to be something other
 * than the correct address.
 */

void
nwlayer::ChangeAddress(frame *fr)
{
	cout << "Changing the address byte\n";
	// Change the address, Form the frame again.
	unsigned char cmd = cmdAddress;
	unsigned char res = resAddress;
	cmdAddress = resAddress = errorAddress;

	// Redo the frame..it will have a wrong address
	// field
	fr->FormFrame(fr->IsCmd());

	// Reset the addresses
	cmdAddress = cmd;
	resAddress = res;
}

