/*
 * File : phlayer.C
 *
 * Description : This file contains the method definitions for the
 *	phlayer class.
 */

#include <string.h>
#include <memory.h>
#include <fcntl.h>
#include <osfcn.h>
#include "scheduler.h"
#include "phlayer.h"
#include "trace.h"
#include "dataPacket.h"
#include "errorCodes.h"
#include "phInactive.h"

DEFINE_TRACE_MODULE(phlayerModule,"phlayerModule");

/*
 * Function : phlayer()
 *
 * Description : This is the constructor. It initializes data
 *	members.
 */


phlayer::phlayer( char *xmitLink, char *recvLink ) :
	inputHandler(-1),
	_pktToSend(0), 
	_lastRecdPkt(0),
	_xmitfd(-1), 
	_recvfd(-1)
{
	_xmitLinkName = new char[strlen(xmitLink)+1];
	strcpy(_xmitLinkName,xmitLink);

	_recvLinkName = new char[strlen(recvLink)+1];
	strcpy(_recvLinkName,recvLink);

}
/*
 * Function : ~phlayer()
 *
 * Description : This is the destructor. Frees memory and
 *	closes the link.
 */

phlayer::~phlayer()
{
	if ( _sm )
		delete _sm;
	CloseLink();
	delete _xmitLinkName;
	delete _recvLinkName;
}

/*
 * Function : Command(int cmd, int &errorCode)
 *
 * Description : Redefinition of the function to handle primitives.
 *	Calls ProcessEvent with the appropriate Event Code.	
 */

void 
phlayer::Command(int cmd, int & errorCode)
{
	TRACE(phlayerModule,1,("Entering Command_int\n"));

	switch ( cmd ) 
		{
	case PH_ActivateReq:
		ProcessEvent(PH_ActivateReq,errorCode);
		break;
	case PH_DeactivateReq:
		ProcessEvent(PH_DeactivateReq,errorCode);
		break;
	default:
		ProcessEvent(PH_Unknown,errorCode);
		break;
		}

	TRACE(phlayerModule,1,("Exiting Command_int\n"));

}

/*
 * Function : Command(int, int &, int)
 *
 * Description : Calls base class function for default functionality.
 */

void 
phlayer::Command(int cmd, int & errorCode, int reason)
{
	layer::Command(cmd,errorCode,reason);
}

/*
 * Function : Command(int, int &, dataPacket *)
 *
 * Description : Redefinition of function to handle Data Request
 *		primitives. First registers with the dataPacket, 
 *		initializes _pktToSend data member, calls ProcessEvent
 *		and then deregisters itself with the dataPacket.
 */

void 
phlayer::Command(int cmd, int &errorCode, dataPacket *data)
{
	TRACE(phlayerModule,1,("Entering Command_int_dataPacket*\n"));


	_pktToSend = data;
	_pktToSend->Register();

	switch ( cmd ) {
		case PH_DataReq:
			TRACE(phlayerModule,2,("Command - PH_DataReq\n"));
			ProcessEvent(PH_DataReq,errorCode);
			break;

		default:
			TRACE(phlayerModule,2,("Unknown command : %d\n",cmd));
			ProcessEvent(PH_Unknown,errorCode);
			break;
	}

	_pktToSend->Deregister();
	_pktToSend = 0;

	TRACE(phlayerModule,1,("Exiting phlayer::Command_int_dataPacket*\n"));
}

/*
 * Function : Init()
 *
 * Description : Init first creates a current state object - so that
 *	the layer is in phInactive state. Then calls Initialize().
 */


int
phlayer::Init()
{
	TRACE(phlayerModule,1,("In Init\n"));
	_sm = new phInactive(this);
	return(Initialize());
}

/*
 * Function : Initialize()
 *
 * Description : Creates FIFOs for transmitting and receiving, then
 *	opens the FIFOs for reading and writing. Finally it adds the
 *	phlayer (this) to the scheduler as an event handler for the
 *	receive FIFO.
 */

int 
phlayer::Initialize()
{
	/*
	 * Make the FIFO for receiveing. 
	 */

	if ( mknod(_recvLinkName, S_IFIFO | 0666, 0) == -1 
								&& errno != EEXIST ) {
		perror("mknod failed");
		return -1;
	}

	TRACE(phlayerModule,1,("Mknod on link %s succeeded\n",
								_recvLinkName));
	/*
	 * Make the FIFO for transmitting. Let it block this 
	 * time, so as to ensure a receiving end has been 
	 * created.
	 */

	if ( mknod(_xmitLinkName, S_IFIFO | 0666, 0) == -1 
								&& errno != EEXIST ) {
		perror("mknod failed");
		return -1;
	}

	TRACE(phlayerModule,1,("Mknod on link %s succeeded\n",
								_xmitLinkName));

	/*
	 * Now open the FIFO's. Use O_NDELAY so as not to
	 * wait for transmitting end.
	 */

	if ( (_recvfd = OpenLink(_recvLinkName,O_RDONLY | 
								O_NDELAY)) == -1 ) {
		return -1;
	}

	TRACE(phlayerModule,1,("OpenLink on link %s succeeded\n",
								_recvLinkName));



	if ( (_xmitfd = OpenLink(_xmitLinkName,O_WRONLY)) 
								== -1 ) {
		return -1;
	}
	TRACE(phlayerModule,1,("OpenLink on link %s succeeded\n",
								_xmitLinkName));

	/*
	 * Now start monitoring input by adding this 
	 * to the scheduler.
	 */

	_fd = _recvfd; // Now monitor input on the receiving FIFO

	scheduler *schd = scheduler::GetScheduler();
	schd->AddEventHandler(this);

	return 0; // Successful
}

/*
 * Function : OpenLink()
 *
 * Description :
 *	Opens a FIFO with the flags.
 */

int 
phlayer::OpenLink(char *link, int flag)
{
	int fd;
	if ( (fd = open(link, flag)) == -1 ) {
		perror("Error opening link");
		return -1;
	}
	else
		return fd;
}

/*
 * Function : EventCallback
 *
 * Description : This method is called by the scheduler when there
 *	is data on the receive FIFO. The data is read, and analyzed
 *	for type of event that has occurred. Then ProcessEvent is
 *	called to handle the event. _lastRecdPkt data member is initialized
 *	to point to the last dataPacket received. 
 */

void
phlayer::EventCallback()
{
	unsigned char buf[MAXLINE];
	int event;
	int bytes;

	TRACE(phlayerModule,1,("\nEntering phlayer::EventCallback\n"));
	
	if ( read(_recvfd,&bytes,sizeof(int))  != sizeof(int)) {
		TRACE(phlayerModule,1,("Error in phlayer::read\n"));
		perror("");
		return;
	}

	TRACE(phlayerModule,1,("Reading %d Bytes\n",bytes));
	int rbytes;
	if ( (rbytes = read(_recvfd,buf,bytes)) != bytes ) {
		TRACE(phlayerModule,1,("Error in phlayer::read : %d\n",rbytes));
		perror("");
		return;
	}
	/*
	 * Create a dataPacket and set _lastRecdPkt to point to it.
	 * Also, register with the dataPacket.
	 */

	_lastRecdPkt = new dataPacket(buf,bytes);
	_lastRecdPkt->Register();

	if ( _lastRecdPkt->Length() == 1 ) {
		unsigned char c = *(_lastRecdPkt->GetData());

		if ( c == PH_ACTIVATE ) {
			TRACE(phlayerModule,1,("Command - PH_ACTIVATE\n"));
			event = PH_Activate;
		}
		else if ( c == PH_DEACTIVATE ) {
			TRACE(phlayerModule,1,("Command - PH_DEACTIVATE\n"));
			event = PH_Deactivate; 
		}
		else {
			event = PH_Unknown;
			TRACE(phlayerModule,1,("Unknown command\n"));
		}
	}
	else {
		TRACE(phlayerModule,1,("Command - PH_DATA\n"));
		event = PH_Data;
	}

	int errorCode = codeSuccessful;
	if ( _blockRecv == FALSE )
		ProcessEvent(event,errorCode);

	// check errorCode here and maybe log it

	_lastRecdPkt->Deregister();
	_lastRecdPkt = 0;

	TRACE(phlayerModule,1,("Exiting phlayer::EventCallback\n"));
}

/*
 * Function : DataRequest
 *
 * Description : Convenience function to transmit a dataPacket
 *	over the xmit FIFO. Transmits an integer equal in value to
 *	the length of the dataPacket, before transmitting the dataPacket.
 */


int
phlayer::DataRequest(dataPacket *data)
{
	unsigned char lenBuf[4];
	int *intPtr;

	if ( _blockXmit == TRUE )
		return 0; // Drop the transmission

	/*
	 * First, append an integer to the dataPacket whose value 
	 * equals the length of the dataPacket. This implements 
	 * messaging on the FIFO.
	 */

	intPtr = (int *)lenBuf;
	*intPtr = data->Length();

	data->Prepend(lenBuf,sizeof(int));
	TRACE(phlayerModule,1,("Header : %d\n",*intPtr));

	/*
	 * Now write it out.
	 */

	TRACE(phlayerModule,1,("Writing %d bytes to fd %d\n",
				data->Length(),_xmitfd));
	if ( write(_xmitfd,data->GetData(),data->Length()) != data->Length() ) {
		perror("error transmitting data");
		return -1;
	}

	/*
	 * Strip the integer that was added.
	 */
	data->PreStrip(sizeof(int));
	return 0;
}

/*
 * Function : SendCommand()
 *
 * Description : This method transmits a single character.
 */

int
phlayer::SendCommand(unsigned char cmd)
{
	TRACE(phlayerModule,1,("Sending char command %c\n",cmd));
	dataPacket *pkt = new dataPacket(&cmd,1);
	pkt->Register();
	int retVal = DataRequest(pkt);	
	pkt->Deregister();
	return retVal;
}

/*
 * Function : CloseLink()
 *
 * Description : Closes the receive and transmit FIFOs.
 *		Removes the phlayer from the scheduler.
 */

void 
phlayer::CloseLink()
{
	TRACE(phlayerModule,1,("Closing links\n"));
	close(_xmitfd);
	close(_recvfd);
	_xmitfd = _recvfd = -1;

	scheduler *schd = scheduler::GetScheduler();
	schd->DeleteEventHandler(this);
}

/*
 * Function : ProcessEvent
 *
 * Description : This is the event handler for the state machine.
 *	It figures out what event has occurred, and calls the appropriate
 *	method on the current state object.
 */

void	
phlayer::ProcessEvent(int event, int &errorCode)
{
	TRACE(phlayerModule,2,("Entering ProcessEvent : %x %d\n",
				_sm,event));
	_sm->PrintState();
	switch ( event ) 
		{
	case PH_ActivateReq:
		_sm = _sm->PhActivateReq(errorCode);
		break;
	case PH_DataReq:
		_sm = _sm->PhDataReq(errorCode);
		break;
	case PH_DeactivateReq:
		_sm = _sm->PhDeactivateReq(errorCode);
		break;
	case PH_Activate:
		_sm = _sm->PhActivate(errorCode);
		break;
	case PH_Data:
		_sm = _sm->PhData(errorCode);
		break;
	case PH_Deactivate:
		_sm = _sm->PhDeactivate(errorCode);
		break;
	default:
		errorCode = codeGeneralFailure;
		TRACE(phlayerModule,2,("Unknown Event\n"));
		break;
		}
	_sm->PrintState();
	TRACE(phlayerModule,2,("Exiting ProcessEvent : %x\n",_sm));
}
